function [FEM]=fem_2d_ad(model,problem_data,parameters)
% FEM_2D_AD Computes matrices and rhs for GaLS-FEM-P1 approximation of
%  -mu Delta u+ b . nabla u +sigma u=f in Omega (2D domain)
%  u=g_D Dirichlet conditions on the boundary.
%  GaLS=Galerkin-Least-Squares stabilization 
%
% [FEM]=fem_2d_ad(model,problem_data,parameters)
%
% Input:
%        model = model object generated by MATLAB (it must contain the
%        field model.Mesh)
%        problem_data = struct with fields:
%          .mu = coefficient (constant) of 2nd order term
%          .b = velocity field (constant) of 1st order term
%          .sigma = coefficient (constant) of 0th order term
%          .f = @(x,y) function-handle of the source
%          .gd = @(x,y) function-handle of Dirichlet condition
%        parameters = struct with fields
%           .degree = degree of exactness of quadrature formulas
%                     (default value=4)   possible values: 1, 3, 4, 10
%           .delta = coefficient for the stabilization
%                     (default value =1)
% Output: 
% 
%         FEM = struct with fields:
%          .Kc:   % matrix A0, (i,j) non-dirichlet
%          .Fc:   % rhs f0, (i) non-dirichlet
%          .B:    % matrix to remap u0 in u
%          .ud:   % ud = discrete lifting Rgd^h of gD
%          .M:    % mass matrix
%

%   Written by Paola Gervasio
%   $Date: 2020/05/14$

if ~isfield(parameters,'degree')
  degree=4;
else
  degree=parameters.degree;
end
if ~isfield(parameters,'delta')
delta=1;
else
  delta=parameters.delta;
end
% extract the mesh from model
 [p,e,t]=meshToPet(model.Mesh);
vmesh.nodes=p';
vmesh.femtype=2;
if strcmp(model.Mesh.GeometricOrder,'linear')
vmesh.p=1;
else
warning('quadratic element not implemented')
return
end
 vmesh.nov=t(1:vmesh.p*3,:);
 vmesh.ne=size(t,2);
 vmesh.np=length(p);
 eb=e(1:2,:);
 vmesh.dof.dir=unique(eb(:));
 vmesh.dof.nondir=setdiff((1:vmesh.np)',vmesh.dof.dir);
% evaluate the basis functions at quadrature nodes
% evaluate elementary matrices  
% evaluate jacobian for each triangle  
[basis,jacf,qf]=ground_zero_fem_2d(vmesh,degree) ;

% build matrices and rhs
[FEM]=fem_2d_mat_ad(vmesh,basis,jacf,qf,problem_data,delta);
    
end   
% -----------------------------------------------------
function [basis,jacf,qf]=ground_zero_fem_2d(vmesh,degree)
% GROUND_ZERO_FEM_2d: functions, jacobians, normal versors
% this function works on all the elements of the domain
%  [basis,jacf,qf]=ground_zero_fem_2d(vmesh,degree,qtype) 
% Input: vmesh = structure with the mesh on the domain
%              .femtype =2 P
%              .p = local polynomial degree (1 or 2)
%              .nodes (nodes of the mesh)
%              .nov (connectivity matrix of the mesh)
%              .ne =  number of elements
%        degree: precision degree of quadrature formulas 
% Output: 
%         basis: structure with fields
%          .phi phi_j(x_q) Lagrange basis at the quadrature nodes
%          .dphix dphi_j/dx(x_q) dx of the Lagrange basis at quad nodes
%          .dphiy dphi_j/dy(x_q) dy of the Lagrange basis at quad nodes
%         jac: cell 
%         jacf{ie}, with ie=1,...ne structure with fields:
%                 .jac = det(J_F_ie) determinant of the Jacobian of the map
%                 .Jmt(2,2) = J^{-T} matrix (used to build the stiffness
%                 matrix) (it is costant on the element)
%                 .J(2,2) = J_F_ie matrix (used to build the stiffness
%                 matrix) (it is costant on the element)
%         qf = structure with fields
%          .x, .w quadrature nodes and weights
% (optional output) boundary =  structure with information on the boundary
%          NOTE:  if this output is required, then the function
%                 boundary_fem_2d.m is called and vmesh.edges is needed
% 
%   Written by Paola Gervasio
%   $Date: 2019/04/11$


% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% compute quadrature nodes and weights in the 2d reference simplex
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


[qf]=quadrature_simplex_2d(degree);



% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% basis functions and its derivatives at the quadrature nodes in the
% reference simplex
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    if vmesh.p==1 % P1
    [basis]=basis_p1_2d(qf);
    elseif vmesh.p==2 % P2
    [basis]=basis_p2_2d(qf);
    end    


% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% basis functions and derivatives at the quadrature nodes in the
% Jacobian matrices
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
jacf=cell(vmesh.ne,1);

for ie=1:vmesh.ne
nov_l=vmesh.nov(:,ie);
jacf{ie}=cosjac_simplex_2d(vmesh.nodes(nov_l,:));
end

% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% compute quadrature nodes and weights in each element of the mesh
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
[nq,~]=size(qf.xq);
ne=vmesh.ne;
qf.xq_ph=zeros(nq,ne);
qf.yq_ph=zeros(nq,ne);
qf.wq_ph=zeros(nq,ne);


for ie=1:vmesh.ne
  novl=vmesh.nov(:,ie);
  V=vmesh.nodes(novl(1),1:2);
  B=jacf{ie}.J;
  jac=jacf{ie}.jac;
X=[qf.xq';qf.yq'];

X1=B*X+[V(1)*ones(1,nq);V(2)*ones(1,nq)];
xq=X1(1,:)';
yq=X1(2,:)';

qf.xq_ph(:,ie)=xq;
qf.yq_ph(:,ie)=yq;
qf.wq_ph(:,ie)=qf.wq*jac;

end
end


% -----------------------------------------------------
function [qf]=quadrature_simplex_2d(degree)
% quadrature_simplex_2d : quadrature nodes and weights in the 2d simplex
% [qf]=quadrature_simplex_2d(degree);
% Input: degree= precision degree of the quadrature formula
% Ouput: qf.xq, .yq, .zq (coordinates of the nodes)
%        qf.wq weights

if degree==1
wq=1/6*ones(3,1);
xq=[0,1,0]';
yq=[0,0,1]';

elseif degree==3
% degree= 3
wq=1/2*[-0.5625;ones(3,1)*25/48];
xq=[1/3,0.2,0.6,0.2]';
yq=[1/3,0.2,0.2,0.6]';

elseif degree==4
% degree= 4
wq =1/2*[0.223381589678011.*ones(1,3) 0.109951743655322.*ones(1,3)]';
xq = [0.445948490915965,0.108103018168070,0.445948490915965,...
     0.091576213509771,0.816847572980459,0.091576213509771]';
yq = [0.445948490915965,0.445948490915965,0.108103018168070,...
     0.091576213509771,0.091576213509771,0.816847572980459]';

elseif degree==10
% degree= 10
wq = 1/2*[0.090817990382754 0.036725957756467*ones(1,3) 0.045321059435528*ones(1,3) 0.072757916845420*ones(1,6) ...
         0.028327242531057*ones(1,6) 0.009421666963733*ones(1,6)]';
    xq = [1/3,0.485577633383657,0.028844733232685,0.485577633383657,...
       0.109481575485037,0.781036849029926,0.109481575485037,0.307939838764121,...
       0.550352941820999,0.141707219414880,0.550352941820999,0.141707219414880,...
       0.307939838764121,0.246672560639903,0.728323904597411,0.025003534762686,...
       0.728323904597411,0.025003534762686,0.246672560639903,0.066803251012200,...
       0.923655933587500,0.009540815400299,0.923655933587500,0.009540815400299,...
       0.066803251012200]';
    yq = [1/3,0.485577633383657,0.485577633383657,0.028844733232685,...
       0.109481575485037,0.109481575485037,0.781036849029926,0.550352941820999,...
       0.307939838764121,0.550352941820999,0.141707219414880,0.307939838764121,...
       0.141707219414880,0.728323904597411,0.246672560639903,0.728323904597411,...
       0.025003534762686,0.246672560639903,0.025003534762686,0.923655933587500,...
       0.066803251012200,0.923655933587500,0.009540815400299,0.066803251012200,...
       0.009540815400299]';
end
qf.xq=xq;
qf.wq=wq;
qf.yq=yq;
end

% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [basis]=basis_p1_2d(qf)
% basis_p1_2d: computes quadrature nodes and weights and evaluate there 
% the P1-basis on the reference triangle T with vertices
%  (0,0,0) (1,0,0), (0,1,0) 
% [basis]=basis_p1_2d(degree)
%
% Input: qf structure with quadrature nodes
% Output:
%        basis.phi = Lagrange basis functions evaluated at the quadrature nodes 
%        basis.phix, .phiy  = Lagrange basis functions derivatives 
%                        evaluated at the quadrature nodes 
%        basis.phi_wq = array of values:  phi_i(x_q) * w_q
%        basis.phi_ij_wq = array of values: phi_j(x_q) * phi_i(x_q) * w_q
%        basis.mat00_{ij}=(phi_j,phi_i)_T
%        basis.mat1x_{ij}=(dphi_j/dx,phi_i)_T
%        basis.mat1y_{ij}=(dphi_j/dy,phi_i)_T
%        basis.mat11_{ij}=(dphi_j/dx,dphi_i/dx)_T
%        basis.mat12_{ij}=(dphi_j/dx,dphi_i/dy)_T
%        basis.mat12_{ij}=(dphi_j/dx,dphi_i/dz)_T
%        basis.mat22_{ij}=(dphi_j/dy,dphi_i/dy)_T

%   Written by Paola Gervasio
%   $Date: 2015/04/15$
%   $Date: 2019/04/27$

nq=length(qf.xq);
zz=zeros(nq,1); % vettore nullo
uno=ones(nq,1);

% basis functions 
basis.phi=[1-qf.xq-qf.yq, qf.xq, qf.yq];
[~,nb]=size(basis.phi);
% derivates:
% dphi/dx
basis.phix=[-uno,uno,zz];
% dphi/dy
basis.phiy=[-uno,zz,uno];

% phi *wq
basis.phi_wq=zeros(nq,nb);
for i=1:nb
basis.phi_wq(:,i)=basis.phi(:,i).*qf.wq;
end

% phi *wq
basis.phi_ij_wq=zeros(nq,nb,nb);

for i=1:nb
for j=1:nb
basis.phi_ij_wq(:,j,i)=basis.phi(:,i).*basis.phi_wq(:,j);
end
end

% mat00_{ij}=(phi_j,phi_i)_T
basis.mat00=[ 1, 1/2, 1/2; 
              1/2, 1, 1/2 ; 
              1/2, 1/2, 1 ]/12;
            
% mat1x_{ij}=(dphi_j/dx,phi_i)_T 
basis.mat1x=[ 1 -1 0;
             -2  2 0;
             -2  2 0]/12; 

% mat1y_{ij}=(dphi_j/dy,phi_i)_T 
 basis.mat1y=[ 1 0 -1;
             -2  0 2;
             -2  0 2]/12; 


  % mat11_{ij}=(dphi_j/dx,dphi_i/dx)_T
basis.mat11=0.5*[1,-1,0; -1,1,0; 0,0,0];

% mat12_{ij}=(dphi_j/dx,dphi_i/dy)_T
basis.mat12=0.5*[1,-1,0; 0,0,0;-1,1,0];

% mat22_{ij}=(dphi_j/dy,dphi_i/dy)_T
basis.mat22=0.5*[1,0,-1; 0,0,0; -1,0,1];
end

% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [jacf]=cosjac_simplex_2d(nodes)
% COSJAC_SIMPLEX_2d: evaluation of the Jacobian of the map on one element
%
% [jacf]=cosjac_simplex_2d(nodes)
% Input: basis = structure with Lagrange basis functions derivatives at the
%                 LGL nodes
%        nodes = mesh nodes in the element 
% Output: jacf.jac = determinant of J_F
%             .Jmt= J^{-T} matrix =J^cof/det(J) (2x2)
%             .J = J_F matrix (2x2)

%   Written by Paola Gervasio
%   $Date: 2019/04/27
 
x=nodes(:,1); y=nodes(:,2);  

jacf.J=[x(2)-x(1), x(3)-x(1);
y(2)-y(1),y(3)-y(1)];
jacf.jac=det(jacf.J);
jacf.Jmt=(jacf.J')\eye(2);

end

% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% assemble matrix and rhs
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [FEM]=fem_2d_mat_ad(vmesh,basis,jacf,qf,problem_data,delta)

[nb,ne]=size(vmesh.nov);

np=max(max(vmesh.nov));
wq=qf.wq;
nq=length(qf.xq);
Ai=zeros(np*8,1);
Aj=Ai;Ax=Ai;
lastA=0;
Mi=zeros(np*8,1);
Mj=Mi;Mx=Mi;
lastM=0;
f=zeros(np,1);
% loop on the elements

for ie=1:ne
  novl=vmesh.nov(:,ie);
  % compute h_ie
  V=[vmesh.nodes(novl,1),vmesh.nodes(novl,2)];
  d1=norm(V(1,:)-V(2,:));
  d2=norm(V(1,:)-V(3,:));
  d3=norm(V(3,:)-V(2,:));
  hk=max([d1,d2,d3]);
  bn=norm(problem_data.b);
  if bn~=0
    tauk=delta*hk/bn;
  else
    tauk=0;
  end
  
  Jmt=jacf{ie}.Jmt; % J^{-T} of the element
  jac=jacf{ie}.jac; % det(J) of the element
  
  B12=problem_data.b;
  mu=problem_data.mu;
  sigma=problem_data.sigma;
  
  % grad(phi) = Jmt/det(B) * \hat grad (\hat phi)
  %           = B^{-T} * \hat grad (\hat phi)
  % in quanto B^{-T}=Jmt/det(B)
  % Jmt e' la matrice dei cofattori
  Gx=zeros(nq,nb); Gy=Gx;
  for j=1:nb
    gphi1=Jmt*[basis.phix(:,j)';basis.phiy(:,j)']  ;
    Gx(:,j)=gphi1(1,:);
    Gy(:,j)=gphi1(2,:);
    
  end
  
  % local Galerkin matrix
  Al=zeros(nb);
  for i=1:nb
    for j=1:nb
      Al(i,j)=jac*(...
        mu*sum((Gx(:,j).*Gx(:,i)+Gy(:,j).*Gy(:,i)).*wq)... % secondo ordine
        +sum((B12(1)*Gx(:,j)+B12(2)*Gy(:,j)).*basis.phi_wq(:,i))... % primo ordine
        +sum(sigma.*basis.phi_ij_wq(:,i,j))); % ordine zero
    end
  end
  
  
  % local mass matrix
  Ml=basis.mat00*jac;
  
  % local GaLS-stabilization matrix
  Sl=zeros(nb);
  for i=1:nb
    for j=1:nb
      Sl(i,j)=jac*...
        sum((B12(1)*Gx(:,j)+B12(2)*Gy(:,j)+sigma*basis.phi(:,j)).*...
        (B12(1)*Gx(:,i)+B12(2)*Gy(:,i)+sigma*basis.phi(:,i)).*wq);
    end
  end
  % Galerkin+ GaLS-stabilization
  Al=Al+tauk*Sl;
  
  % global assembling
  [Ai,Aj,Ax,lastA]=assemble_triplet(Ai,Aj,Ax,lastA,Al,nb,nb,...
    novl,novl);
  [Mi,Mj,Mx,lastM]=assemble_triplet(Mi,Mj,Mx,lastM,Ml,nb,nb,...
    novl,novl);
  
  % local rhs (Galerkin +GaLS)
  fl=zeros(nb,1);
  for i=1:nb
    fl(i)=jac*sum(problem_data.f(qf.xq_ph(:,ie),qf.yq_ph(:,ie)).*wq.*...
      (basis.phi(:,i)+tauk*(B12(1)*Gx(:,i)+B12(2)*Gy(:,i)+sigma*basis.phi(:,i))));
  end
  % assembling rhs
  f(novl)=f(novl)+fl;
end


Ai=Ai(1:lastA);
Aj=Aj(1:lastA);
Ax=Ax(1:lastA);
Mi=Mi(1:lastM);
Mj=Mj(1:lastM);
Mx=Mx(1:lastM);

% [i,j,A] ---> sparse
A=sparse(Ai,Aj,Ax,np,np);
M=sparse(Mi,Mj,Mx,np,np);

% reduce the matrix A to non-dirichelt dofs
FEM.Kc=A(vmesh.dof.nondir,vmesh.dof.nondir);
FEM.M=M;
% matrix B to remap u0 in the global u
npi=length(vmesh.dof.nondir);
B=zeros(vmesh.np,npi);
for i=1:npi
  B(vmesh.dof.nondir(i),i)=1;
end
FEM.B=B;

% ud is the discrete lifting of gD
ud=zeros(vmesh.np,1);
ud(vmesh.dof.dir)=problem_data.gd(vmesh.nodes(vmesh.dof.dir,1),vmesh.nodes(vmesh.dof.dir,2));
FEM.ud=ud;
%

% rhs reduced to nondirichlet dofs
FEM.Fc=f(vmesh.dof.nondir)-A(vmesh.dof.nondir,vmesh.dof.dir)*ud(vmesh.dof.dir);
end


function [ti,tj,tx,last]=assemble_triplet(ti,tj,tx,last,Al,nrow,ncol,...
novr,novc) 
% assemble triplet: matrix assembling, from local 2-index matrix to 3 arrays
% [ti,tj,tx,last]=assemble_triplet(ti,tj,tx,last,Al,nrow,ncol,novr,novc)
%
% Input: ti, tj, tx = rows, columns, values arrays (of the global matrix)
%        last = last element of ti, tj, tx previously filled in
%        Al = 2-index array (local matrix)
%        nrow, ncol= numero of rows and columns of Al
%        novr, novc= local2global maps for rows and columns
% Output: ti, tj, tx = updated rows, columns, values arrays 
%        last = last element of ti, tj, tx  filled in during this call

%   Written by Paola Gervasio
%   $Date: 2007/04/01$

nt=nrow*ncol;
ig=novr(1:nrow);
jg=novc(1:ncol);

i=zeros(nrow,ncol);
j=zeros(nrow,ncol);
for k=1:ncol
    i(:,k)=ig;
end
for k=1:nrow
    j(k,:)=jg;
end
ti(last+1:last+nt)=i(:); 
tj(last+1:last+nt)=j(:);  
tx(last+1:last+nt)=Al(:);
% 
last=last+nt;
end
